// Simulator.h

#ifndef __Simulator_h__
#define __Simulator_h__

#include "Patch.h"
#include "MersenneTwister.h"
#include "OrgID.h"
#include <functional>
#include <cassert>
#include <vector>
#include <hash_map>
#include <map>
#include <sstream>


// #defines associated with parameter registry
#define _MaxParamFileLineLength 1000
#define _MaxParamNameLength 75
#define _MaxIntParams 50
#define _MaxUIntParams 50
#define _MaxULongParams 20
#define _MaxFloatParams 50
#define _MaxBoolParams 20
#define _MaxStringParams 20
#define _RegisterParamInt(P,N) { IntParams[numIntParams] = (int*)P; strcpy(IntParamNames[numIntParams], N); numIntParams++; }
#define _RegisterParamUInt(P,N) { UIntParams[numUIntParams] = (unsigned int*)P; strcpy(UIntParamNames[numUIntParams], N); numUIntParams++; }
#define _RegisterParamULong(P,N) { ULongParams[numULongParams] = (unsigned long*)P; strcpy(ULongParamNames[numULongParams], N); numULongParams++; }
#define _RegisterParamFloat(P,N) { FloatParams[numFloatParams] = (float*)P; strcpy(FloatParamNames[numFloatParams], N); numFloatParams++; }
#define _RegisterParamBool(P,N) { BoolParams[numBoolParams] = (bool*)P; strcpy(BoolParamNames[numBoolParams], N); numBoolParams++; }
#define _RegisterParamString(P,N) { StringParams[numStringParams] = (char*)P; strcpy(StringParamNames[numStringParams], N); numStringParams++; }

// some more general #defines
#define mymin(A,B) ((A<B) ? A : B)
#define mymax(A,B) ((A>B) ? A : B)
#define mybounded(A) ((A>1.0f) ? 1.0f : ((A < 0.0f) ? 0.0f : A))


class Simulator
{
public:
	Simulator();
	~Simulator();

	void run( int maxIterations );
	
	float getLocalDensity( int x, int y, bool plusOne = false );

	int wx( int x ); // return x wrapped to be in environment bounds
	int wy( int y ); // return y wrapped to be in environment bounds
	int fwx( int x ); // return x wrapped to be in environment bounds, fast version - x must be at most ENV_SIZE_X out of bounds
	int fwy( int y ); // return y wrapped to be in environment bounds, fast version - y must be at most ENV_SIZE_Y out of bounds

	Patch & getPatch( int x, int y );
	Patch & getPatchW( int x, int y );
	Patch & getPatchFW( int x, int y );

	void getNewOrgID( int specificity, int numgenes, OrgID * pNewID );
	unsigned int getOrgIDKey( int specificity, int numgenes ) const;

	void recordOrgBirth( OrgID & id );
	void recordOrgDeath( Organism * pOrg );
	
	bool inPhylogenyRecordWindow();
	
	static void loadParameterValues( const char * paramfilename );
	static void recordParameterValues( ostream& os );
	
	struct SpeciesInfo
	{
		SpeciesInfo() : count(1), numDead(0), cumulativeAgeAtDeath(0), cumulativeNumOffspring(0) {};
		int count;
		int numDead;
		long cumulativeAgeAtDeath;
		long cumulativeNumOffspring;
	};

	enum eDeathPolicy
	{
		dpConstant = 0,
		dpOneMinusFitness = 1,
		dpOneMinusFitnessTimesOneMinusSpec = 2,
		dpOneMinusProbRep = 3,
		dpOneMinusProbRepTimesOneMinusSpec = 4
	};
	
	enum eAbioticEnvType
	{
		aetSimple = 0,
		aetComplex = 1
	};
	
	enum eReproductionFunction
	{
		rfLinear = 0,
		rfPower = 1,
		rfSigmoid = 2
	};
	
	// simulation parameters
	static unsigned long RNG_SEED;
	
	static int NUM_ITERATIONS;
	static int ENV_SIZE_X;
	static int ENV_SIZE_Y;
	static int DISPERSAL_RADIUS;
	static int NEIGHBOURHOOD_RADIUS;
	static eReproductionFunction REPRODUCTION_FUNCTION_TYPE;
	static float REPRODUCTION_FUNCTION_COEFF_K;
	static float REPRODUCTION_FUNCTION_COEFF_C;
	static float REPRODUCTION_FUNCTION_COEFF_X;
	static eDeathPolicy DEATH_POLICY;
	static float DEATH_PROB_COEFF_X;
	static float LOCAL_POP_DENSITY_MAX;
	static float LOCAL_POP_DENSITY_INIT;
	
	static float MUTATION_ADD_PROB;
	static float MUTATION_DELETE_PROB;
	static float MUTATION_CHANGE_PROB_PER_GENE;
	static float MUTATION_CHANGE_PROB_REQ;
	static bool MUTATION_PROPORTIONAL_TO_LENGTH;
	static bool ORGANISM_REQUIREMENTS_EVOLVABLE; 
	static bool ENVIRONMENT_DISTURBANCE_ENABLED;
	static bool ENVIRONMENT_DISTURBANCE_RESTRICTED;
	static int ENVIRONMENT_DISTURBANCE_MAX_SPREAD;
	static float ORGANISM_SPECIFICITY_FRACTION_INIT;

	static int LOCAL_ENVIRONMENT_VECTOR_SIZE;
	static float LOCAL_ENVIRONMENT_VECTOR_INIT_VAL;
	static eAbioticEnvType ABIOTIC_ENV_TYPE;
	
	static int NEW_GENOME_LENGTH_MIN;
	static int NEW_GENOME_LENGTH_MAX;
	
	// data recording parameters
	static bool RECORD_PHYLOGENY_FILE_FULL;
	static bool RECORD_PHYLOGENY_FILE_PARTIAL;
	static bool RECORD_GENOME_BANK_FILE;
	static bool RECORD_GENOME_SIZE_FILE;
	static bool RECORD_GENOME_SPECIFICITY_FILE;
	static bool RECORD_SPECIES_COUNTS_FILE;
	static bool RECORD_SIMULATION_STATE;
	
	static int DATA_RECORD_PERIOD;
	static int SIMULATION_STATE_RECORD_PERIOD;
	static int PHYLOGENY_PARTIAL_RECORD_PERIOD;
	static int PHYLOGENY_PARTIAL_RECORD_WINDOW;
	static int GENOME_SIZE_FILE_MAX_SIZE;
	static int GENOME_BANK_RECORD_THRESHOLD; // num of offspring before a genome is recorded
	static int SPECIES_NUMBER_SIG_THRESHOLD; // for recording diversity of significant species in summary file
	static int SPECIES_COUNTS_FILE_ZIP_PERIOD;
	
	// derived values
	static int NEIGHBOURHOOD_SIZE;

	// miscellaneous constants
	static const float FLOATING_POINT_ACCURACY = 0.000001f;

	static MTRand m_RNG;
	static long NRC_RNG_STATE;
	
	/* function object which can be used with STL algorithms such as random_shuffle */
	struct RNGFunctionObject : public unary_function< int, int>
	{
		int operator()( int n ) { return (Simulator::m_RNG.randInt() % n); }
	};
	
	// parameter registry data
	static int numIntParams;
	static int numUIntParams;
	static int numULongParams;
	static int numFloatParams;
	static int numBoolParams;
	static int numStringParams;
	static int *IntParams[_MaxIntParams];
	static char IntParamNames[_MaxIntParams][_MaxParamNameLength];
	static unsigned int *UIntParams[_MaxUIntParams];
	static char UIntParamNames[_MaxUIntParams][_MaxParamNameLength];
	static unsigned long *ULongParams[_MaxULongParams];
	static char ULongParamNames[_MaxULongParams][_MaxParamNameLength];
	static float *FloatParams[_MaxFloatParams];
	static char FloatParamNames[_MaxFloatParams][_MaxParamNameLength];
	static bool *BoolParams[_MaxBoolParams];
	static char BoolParamNames[_MaxBoolParams][_MaxParamNameLength];
	static char *StringParams[_MaxStringParams];
	static char StringParamNames[_MaxStringParams][_MaxParamNameLength];	


private:
	// methods
	void initialise();
	void checkOvercrowding();
	void recordSummaryData( int iteration );
	void saveState( int iteration );
	void getSpeciesCountsFilename( stringstream & scname, int idx );
	static char * gstrtokfromfile( char * line, char * separators, const char * filename, FILE* fp, char * errormsg );
	static void setDerivedVariables();

	// data members
	vector<Patch> m_Environment;
	hash_map< unsigned int, unsigned int > m_OrgIDMap;				// contains record of all used Organism IDs
	map< OrgID, SpeciesInfo, OrgID::ltOrgID > m_CurrentSpeciesMap;	// contains IDs and counts of all orgs currently alive

	int m_CurrentIteration;
	int m_LatestSpeciesCountsIdx;
};


inline int Simulator::wx( int x )
{
	return ( (x >= 0) ? (x % ENV_SIZE_X) : (ENV_SIZE_X + (x % ENV_SIZE_X)) );
}


inline int Simulator::wy( int y )
{
	return ( (y >= 0) ? (y % ENV_SIZE_Y) : (ENV_SIZE_Y + (y % ENV_SIZE_Y)) );
}


inline int Simulator::fwx( int x )
{
	return ( (x < 0) ? (x + ENV_SIZE_X) : ((x >= ENV_SIZE_X) ? (x - ENV_SIZE_X) : x) );
}


inline int Simulator::fwy( int y )
{
	return ( (y < 0) ? (y + ENV_SIZE_Y) : ((y >= ENV_SIZE_Y) ? (y - ENV_SIZE_Y) : y) );
}


inline Patch & Simulator::getPatch( int x, int y )
{
	//assert(x >= 0);
	//assert(x < ENV_SIZE_X);
	//assert(y >= 0);
	//assert(y < ENV_SIZE_Y);
	return m_Environment[ ENV_SIZE_Y * x + y ];
}


inline Patch & Simulator::getPatchW( int x, int y )
{
	return m_Environment[ ENV_SIZE_Y * wx(x) + wy(y) ];
}


inline Patch & Simulator::getPatchFW( int x, int y )
{
	return m_Environment[ ENV_SIZE_Y * fwx(x) + fwy(y) ];
}


inline unsigned int Simulator::getOrgIDKey( int specificity, int numgenes ) const
{
	return (((unsigned int)numgenes * (unsigned int)LOCAL_ENVIRONMENT_VECTOR_SIZE) + (unsigned int)specificity);
}


#endif
