// actiview.cc
// Process data from speciescounts.dat to generate activity wave data
//
// Usage: actiview t p a 
// where t = threshold conc above which species are included in wave files
//       p = period of recording (i.e. only read 1 line in every p)
//       a = activity theshold above which species are recorded in waveannote.txt
//
// compile with: gcc -Wall actiview.cc OrgID.cc -I../include -lstdc++ -o ../bin/actiview
//

#include "OrgID.h"
#include <fstream>
#include <set>
#include <map>
#include <slist>
#include <string>
#include <sstream>
#include <cstdlib>
#include <cassert>


struct orginfo
{
	orginfo() : activity(0), lastupdate(0) {};
	orginfo( unsigned int act, unsigned int up ) : activity(act), lastupdate(up) {};
	unsigned int activity;
	unsigned int lastupdate;
};


void writeData( map< OrgID, slist<orginfo>, OrgID::ltOrgID > & activitymap, bool writeall, int gen, int * pOrgNum, int a );



int main( int argc, const char *argv[] )
{
	if (argc != 4)
	{
		cerr << "ACTIVIEW usage: actiview t p a" << endl
		<< "where\tt = threshold conc above which species are included in wave files" << endl
		<< "\tp = period of recording (i.e. only read 1 line in every p)" << endl
		<< "\ta = activity theshold above which species are recorded in waveannote.txt" << endl;
		exit( EXIT_FAILURE );
	}
	
	int t = atoi(argv[1]);
	int p = atoi(argv[2]);
	int a = atoi(argv[3]);
	
	ifstream concfile( "speciescounts.dat" );
	
	if (concfile.bad())
	{
		cerr << "Could not open speciescounts.dat file!" << endl;
		exit( EXIT_FAILURE );
	}
	
	ofstream sigdiv( "sigdiversity.dat" );
	
	if (! sigdiv.good())
	{
		cerr << "Could not open sigdiversity.dat file for writing!" << endl;
		exit( EXIT_FAILURE );
	}
	
	// first pass of the file
	// record all species with counts above threshold t into sigspecies set
	
	set< OrgID, OrgID::ltOrgID > sigspecies;
	string lines;
	OrgID id;
	int linenum = 1;
	int gen, count;
	
	while ( getline(concfile, lines) )  // this version of getline defined in <string>
	{
		if ( linenum % p == 0 )
		{
			stringstream liness( lines );
			liness >> gen;
			if (!liness)
			{
				cout << "Error reading file speciescounts.dat, line " << linenum << endl;
				exit( EXIT_FAILURE );
			}
	
			cout << "Generation: " << gen << endl;
			
			int sigdiversity = 0;
	
			while (! liness.eof() )
			{
				liness >> id >> count;
				if (count >= t)
				{
					sigspecies.insert(id);
					++sigdiversity;
				}
			}
			
			sigdiv << gen << " " << sigdiversity << endl;
		}

		++linenum;
	}
	
	sigdiv.close();
	
	// second pass of file
	// record cumulative activity of any species in sigspecies
	
	concfile.close();
	concfile.open( "speciescounts.dat" );
	if (concfile.bad())
	{
		cerr << "Could not open speciescounts.dat file!" << endl;
		exit( EXIT_FAILURE );
	}	

	
	map< OrgID, slist<orginfo>, OrgID::ltOrgID > activitymap;
	
	linenum = 1;
	
	int orgnum = 0;
	
	while ( getline(concfile, lines) )  // this version of getline defined in <string>
	{
		if ( linenum % p == 0 )
		{
			stringstream liness( lines );
			liness >> gen;
			if (!liness)
			{
				cout << "Error reading file speciescounts.dat, line " << linenum << endl;
				exit( EXIT_FAILURE );
			}
	
			cout << "Generation: " << gen << endl;
	
			while (! liness.eof() )
			{
				liness >> id >> count;
				set< OrgID, OrgID::ltOrgID >::iterator spec_it = sigspecies.find(id);
				if (spec_it != sigspecies.end())
				{
					map< OrgID, slist<orginfo>, OrgID::ltOrgID >::iterator act_it = activitymap.find(id);
					if (act_it != activitymap.end())
					{
						orginfo info( count + act_it->second.front().activity, gen );					
						act_it->second.push_front( info );
					}
					else
					{
						orginfo info( count, gen );					
						slist<orginfo> vec;
						vec.push_front( info );
						activitymap.insert( make_pair(id,vec) );
					}
				}
			}
			
			// now go through activitymap and write data to wave.dat and waveannote.txt files
			writeData( activitymap, false, gen, &orgnum, a );
		}

		++linenum;
	}
	
	// now go through activitymap and write any remaining data to wave.dat and waveannote.txt files
	//writeData( activitymap, true, 0, &orgnum, a );
	writeData( activitymap, false, gen+1, &orgnum, a );
	
	
	ofstream wavinf( "waveinfo.dat" );
	if (wavinf.good())
		wavinf << orgnum << endl;
	wavinf.close();
	
	assert( activitymap.empty() );
	
	return EXIT_SUCCESS;
}


void writeData( map< OrgID, slist<orginfo>, OrgID::ltOrgID > & activitymap, bool writeall, int gen, int * pOrgNum, int a )
{
	map< OrgID, slist<orginfo>, OrgID::ltOrgID >::iterator act_it  = activitymap.begin();
	map< OrgID, slist<orginfo>, OrgID::ltOrgID >::iterator act_end = activitymap.end();
	
	while (act_it != act_end)
	{
		if ( writeall || ((act_it->second.front()).lastupdate < (unsigned int)gen ))
		{
			// species no longer present, so record and delete it
			stringstream wavname;
			wavname << "wave" << *pOrgNum << ".dat";
			ofstream wav(wavname.str().c_str(), ios::app);
			if (!wav.good())
			{
				cerr << "Could not open wave.dat file for writing!" << endl;
				exit( EXIT_FAILURE );
			}
			
			slist<orginfo> & data = act_it->second;
			slist<orginfo>::iterator data_it  = data.begin();
			slist<orginfo>::iterator data_end = data.end();
			unsigned int maxactivity = 0;
			while (data_it != data_end)
			{
				wav << data_it->lastupdate << " " << data_it->activity << " ";
				if (data_it->activity > maxactivity)
					maxactivity = data_it->activity;
				++data_it;
			}
			wav << endl;
			wav.close();
			
			if (maxactivity >= (unsigned int)a)
			{
				stringstream wavannname;
				wavannname << "waveannote" << *pOrgNum << ".txt";
				ofstream wavann(wavannname.str().c_str(), ios::app);
				if (!wavann.good())
				{
					cerr << "Could not open waveannote.txt file for writing!" << endl;
					exit( EXIT_FAILURE );
				}
				wavann << act_it->first << endl;
				wavann.close();			
			}
			
			map< OrgID, slist<orginfo>, OrgID::ltOrgID >::iterator act_next = act_it;
			++act_next;
			activitymap.erase( act_it );
			act_it = act_next;
			
			(*pOrgNum)++;
		}
		else
		{
			++act_it;
		}
	}	
}
