// analysepatch.cc
// Process data from a specified patch in a niche-state file
//
// Usage: analysepatch file i x y
// where file = name of niche-state file to be read
//       i = iteration number (used for naming output files)
//       x = x coordinate of patch to be analysed
//       y = y coordinate of patch to be analysed
//
// compile with: gcc -Wall -g analysepatch.cc OrgID.cc -I../include -lstdc++ -o ../bin/analysepatch
//

#include "OrgID.h"
#include <fstream>
#include <string>
#include <sstream>


int main( int argc, const char *argv[] )
{
	if (argc != 5)
	{
		cerr << "ANALYSEPATCH usage: analysepatch file i x y" << endl
		<< "where\tfile = name of niche-state file to be read" << endl
		<< "\ti = iteration number (used for naming output files)" << endl
		<< "\tx = x coordinate of patch to be analysed" << endl
		<< "\ty = y coordinate of patch to be analysed" << endl;
		exit( EXIT_FAILURE );
	}
	
	int IT = atoi(argv[2]);
	int X = atoi(argv[3]);
	int Y = atoi(argv[4]);	
	
	ifstream statefile( argv[1] );
	
	if (statefile.bad())
	{
		cerr << "Could not open file " << argv[1] << "!" << endl;
		exit( EXIT_FAILURE );
	}

	// go through state file looking for the specified patch
	string lineStr;
	int linenum = 1;
	bool lineFound = false;
	string coordStr;
	int x, y, commaPos;
	
	while ( getline(statefile, lineStr) && !lineFound )  // this version of getline defined in <string>
	{
		stringstream lineStrStr( lineStr );
		lineStrStr >> coordStr;
		
		if (!lineStrStr)
		{
			cout << "Error reading file " << argv[1] << ", line " << linenum << endl;
			exit( EXIT_FAILURE );
		}

		commaPos = coordStr.find(",",0);
		if (commaPos <= 0)
		{
			cout << "Error reading file " << argv[1] << ", line " << linenum << ". Unable to parse patch coordinates." << endl;
			exit( EXIT_FAILURE );
		}
		
		x = atoi( coordStr.substr(0, commaPos).c_str() );
		y = atoi( coordStr.substr(commaPos+1, coordStr.length() - commaPos - 1).c_str() );
		
		if ((x == X) && (y == Y))
		{
			// found it!

			OrgID orgID, parentID;
			int numOrgs, timeOfBirth, numOffspring, specificity, reqStartPos, numGenes, func, part, spread, repeat, op;
			float specFraction, fitness, probDeath, probReproduction, fVal;

			lineStrStr >> numOrgs;

			cout << "Line " << linenum << ", x=" << x << ", y=" << y << ", number of organisms=" << numOrgs << endl;
			
			stringstream repFilename;
			repFilename << "report_" << IT << "_" << X << "_" << Y << ".txt";
			ofstream repFile( repFilename.str().c_str() );
			
			repFile << "REPORT FOR PATCH " << X << "," << Y << " OF " << argv[1] << endl;
			repFile << "Number of Organisms: " << numOrgs << endl;
			repFile << "Summary of Organisms =======================================" << endl;

			for (int org = 0; org < numOrgs; ++org)
			{
				lineStrStr	>> orgID >> parentID >> timeOfBirth >> numOffspring
							>> specificity >> specFraction >> reqStartPos >> fitness
							>> probDeath >> probReproduction >> numGenes;
				
				for (int i = 0; i < numGenes; ++i)
				{
					lineStrStr >> func >> part >> spread >> repeat >> op;
				}
				
				// read in Env Disturbance (gene)
				lineStrStr >> func >> part >> spread >> repeat >> op;
				
				repFile << "- Organism " << (org+1) << " -----------------" << endl;
				repFile << "   ID=" << orgID << ", Parent ID=" << parentID << ", Number of Genes=" << numGenes << endl;
				repFile << "   Time of Birth=" << timeOfBirth << ", Number of Offspring=" << numOffspring << endl;
				repFile << "   Specificity=" << specificity << " (spec frac=" << specFraction << "), Requirements Start Pos=" << reqStartPos << endl;
				repFile << "   Fitness=" << fitness << ", Probability of Death=" << probDeath << ", Probability of Reproduction=" << probReproduction << endl;
				
				// read in Phenotype
				int phSize;
				stringstream phFilename;
				phFilename << "phenotype_" << orgID << "_" << IT << "_" << X << "_" << Y << ".dat";
				ofstream phFile( phFilename.str().c_str() );
				lineStrStr >> phSize;
				for (int j = 0; j < phSize; ++j)
				{
					lineStrStr >> fVal;
					phFile << fVal << " ";
				}
				phFile.close();				
			}
			
			// read in local abiotic env
			int laeSize;
			stringstream laeFilename;
			laeFilename << "localAbioticEnv_" << IT << "_" << X << "_" << Y << ".dat";
			ofstream laeFile( laeFilename.str().c_str() );
			lineStrStr >> laeSize;
			for (int i = 0; i < laeSize; ++i)
			{
				lineStrStr >> fVal;
				laeFile << fVal << " ";
			}
			laeFile.close();
			
			// read in current modified env
			int cmeSize;
			stringstream cmeFilename;
			cmeFilename << "currentModifiedEnv_" << IT << "_" << X << "_" << Y << ".dat";
			ofstream cmeFile( cmeFilename.str().c_str() );
			lineStrStr >> cmeSize;
			for (int i = 0; i < cmeSize; ++i)
			{
				lineStrStr >> fVal;
				cmeFile << fVal << " ";
			}
			cmeFile.close();
			
			lineFound = true;
		}

		++linenum;
	}
	
	statefile.close();

	if (!lineFound)
	{
		cerr << "Unable to find data for patch " << X << "," << Y << " in file " << argv[1] << "!" << endl;
		exit( EXIT_FAILURE );
	}
	
	return EXIT_SUCCESS;
}
